// BImage.cpp (this is -*- C++ -*-)
// 
// \author: Bjoern Giesler <bjoern@giesler.de>
// 
// 
// 
// $Author: giesler $
// $Locker$
// $Revision$
// $Date: 2002-08-19 10:41:28 +0200 (Mon, 19 Aug 2002) $

/* system includes */
#include <string.h>
#include <stdlib.h>
#include <nds/arm9/video.h>

/* my includes */
#include "BImage.h"
#include "BController.h"
#include "BLogger.h"

BImage::BImage(int x, int y, char** xpmData):
  BDrawable(x, y)
{
  long ncolors, chars_per_pixel;
  char *fmtString = xpmData[0];
  char *tmp = fmtString;
  int len = strlen(fmtString);

  _width = (int)strtol(tmp, &tmp, 10);
  tmp++; // skip space
  if(tmp - fmtString > len)
    {
      LOG(LOG_WARN, "At end of format string '%s' after width\n", fmtString);
      return;
    }

  _height = (int)strtol(tmp, &tmp, 10);
  tmp++; // skip space
  if(tmp - fmtString > len)
    {
      LOG(LOG_WARN, "At end of format string '%s' after height\n", fmtString);
      return;
    }

  ncolors = strtol(tmp, &tmp, 10);
  tmp++; // skip space
  if(tmp - fmtString > len)
    {
      LOG(LOG_WARN, "At end of format string '%s' after ncolors\n", fmtString);
      return;
    }

  chars_per_pixel = strtol(tmp, &tmp, 10);
  if(tmp - fmtString != len)
    {
      LOG(LOG_WARN, "Stuff at end of format string '%s' after chars per pixel\n", fmtString);
      return;
    }

  uint16* colortable = new uint16[ncolors];
  char **colornames = new char *[ncolors];
  for(int i=0; i<ncolors; i++)
    {
      char* colorline = xpmData[i+1];
      colornames[i] = colorline;
      char* color = colorline + chars_per_pixel + 3;
      if(color - colorline > (int)strlen(colorline))
	{
	  LOG(LOG_WARN, "Invalid color line %d: '%s' (%s)\n", i, colorline, color);
	  return;
	}

      if(strcmp(color, "None") == 0)
	colortable[i] = 1<<15; 
      else if(color[0] == '#')
	{
	  long color24 = strtol(color+1, NULL, 16);
	  char r = (color24 & 0x00ff0000) >> 16;
	  char g = (color24 & 0x0000ff00) >> 8;
	  char b = (color24 & 0x000000ff);
	  r >>= 3;
	  g >>= 3;
	  b >>= 3;
	  colortable[i] = RGB15(r, g, b);
	}
      else
	{
	  LOG(LOG_WARN, "Invalid color line %d: %s\n", i, colorline);
	  return;
	}				
    }

  _image = new uint8[_width*_height*2];
  _format = FMT_RGB15;
  _freeImage = true;
  
  for(int row=0; row<_height; row++)
    {
      char* line = xpmData[1+ncolors+row];
      for(int col=0; col<_width; col++)
	{
	  char* colorname = &(line[col*chars_per_pixel]);
	  bool colorfound = false;
	  for(int i=0; i<ncolors; i++)
	    {
	      if(strncmp(colorname, colornames[i], chars_per_pixel) == 0)
		{
		  _image[row*_width+col] = colortable[i];
		  colorfound = true;
		  break;
		}
	    }

	  if(!colorfound)
	    {
	      LOG(LOG_WARN, "No match for color '%s'\n", colorname);
	      return;
	    }
	}
    }

  delete colortable;
  delete colornames;
}

BImage::BImage(int x, int y, int width, int height,
	       uint8* image, ImageFormat format, bool copy):
  BDrawable(x, y)
{
  if(copy)
    {
      if(format == FMT_RGB15)
	{
	  _image = new uint8[width*height*2];
	  memcpy(_image, image, width*height*2);
	}
      else
	{
	  _image = new uint8[width*height];
	  memcpy(_image, image, width*height);
	}
	
      _freeImage = true;
    }
  else
    {
      _image = image;
      _freeImage = false;
    }
  _format = format;
  _width = width;
  _height = height;
}

BImage::BImage(const BImage& img):
  BDrawable(img._x, img._y)
{
  _width = img._width;
  _height = img._height;
  _format = img._format;
  if(_format == FMT_RGB15)
    {
      _image = new uint8[_width*_height*2];
      memcpy(_image, img._image, _width*_height*2);
    }
  else
    {
      _image = new uint8[_width*_height];
      memcpy(_image, img._image, _width*_height);
    }
  _freeImage = true;
}

BImage::~BImage()
{
  if(_freeImage)
    delete _image;
}

void
BImage::operator=(const BImage& img)
{
  if(_freeImage) delete _image;
  
  _width = img._width;
  _height = img._height;
  _format = img._format;
  
  if(_format == FMT_RGB15)
    {
      _image = new uint8[_width*_height*2];
      memcpy(_image, img._image, _width*_height*2);
    }
  else
    {
      _image = new uint8[_width*_height];
      memcpy(_image, img._image, _width*_height);
    }
  _freeImage = true;
}

void BImage::draw(uint16* screen)
{
  if(_x > SCREEN_WIDTH || _x + _width < 0 ||
     _y > SCREEN_HEIGHT || _y + _height < 0 ||
     _format != FMT_RGB15)
    // nothing to draw.
    return;

  int x = (_x < 0 ? 0 : _x);
  int y = (_y < 0 ? 0 : _y);
  int xoffs = 0, yoffs = 0;
  int xlen = _width, ylen = _height;
  if(_x<0)
    {
      xoffs = -_x;
      xlen -= xoffs;
    }
  if(_y<0)
    {
      yoffs = -_y;
      ylen -= yoffs;
    }
  if(_x + xlen > SCREEN_WIDTH)
    xlen = SCREEN_WIDTH - _x;
  if(_y + ylen > SCREEN_HEIGHT)
    ylen = SCREEN_HEIGHT - _y;

  //  uint16 *img = &(_image[_width]);
  uint16 *img = (uint16*)_image;
  for(int row = 0; row<_height; row++)
    {
      memcpy(&(screen[(y+row)*SCREEN_WIDTH+x]),
	     &(img[(row+yoffs)*_width + xoffs]),
	     2*xlen);
    }
}
